#include<sys/types.h>
#include<sys/time.h>
#include<sys/wait.h>
#include<unistd.h>
#include<stdio.h>
#include<stdlib.h>
#include<strings.h>
#include<syslog.h>
#include<time.h>
#include<signal.h>
#include<sys/resource.h>
#include<fcntl.h>

#include "./config.h"

/********************************************************************************/
/*            ALL THIS SETTINGS MUST BE GREATER THAN ZERO 			*/

/* above this CPU usage (percentage) limits processes will be considered bad */
#define CPU_LEVEL1 80
#define CPU_LEVEL2 10

/* if average system load is above the respective values */
#define SYS_LOAD_LEVEL1 200
#define SYS_LOAD_LEVEL2 400


/* above this MEM usage (percentage) limite processes will be considered bad, this is
independent from system average load */
#define MEM_LIMIT 30 

/* consecutive main failures for system reboot */
#define MAIN_FAIL_REBOOT 8

/* if system average load is above this value it is considered to be a major fault */
#define MAIN_FAIL_LOAD_LIMIT 10000

/* cycle checking delay, should be greater than 4 seconds */
#define CICLE_DELAY 15

/* consecutive times a process must be considered bad to be generated a warning */
#define COUNT_TO_WARN 1 

/* consecutive times a process must be considered bad to be placed in low priority */
#define COUNT_TO_NICE 2

/* consecutive times a process must be considered bad to send a TERM signal */
#define COUNT_TO_TERM 6

/* consecutive times a process must be considered bad to send a KILL signal */
#define COUNT_TO_KILL 7

/* check root processes too */
#define CHECK_ROOT 0

/********************************************************************************/

#define PS_ARGS "xauhc"

#define MAX_BAD_PIDS 200



/* constant for maxCpu calc */
int kkkk=(1000*CPU_LEVEL1-1000*CPU_LEVEL2)/(SYS_LOAD_LEVEL2-SYS_LOAD_LEVEL1);




/* safe users in command line args, root is always safe ,
this means root owned processes are never checked */

int psPid;
char reading, mainFail;

int readLine(char *buff, int buffLen, int fd)
{
char *aux=buff;
int i, res;
for(;;)
	{
	i=5;
	do 
		{
		res=read(fd,aux,1); 
		if(res!=1 && i--) sleep(1);
		}
	while(res!=1 && i);
	if(res==1) 
		{
		if(*aux=='\n' || aux-buff==buffLen) {*aux=0; return(1);}
		else aux++;
		}
	else return(0);
	}
	
}

void stopPs(int a)
{
if(reading && psPid>0) {kill(psPid,SIGKILL);mainFail=1;}
signal(SIGALRM,stopPs);
}

void waitChild(int a)
{
int i;
wait(&i);
signal(SIGCHLD,waitChild);
}


void main(int argc, char *argv[])
{
char line[303];
char mainFailCount, issys, *aux, *aux1;
int i, myPpid, myPid, pid, load, mem, sysLoad;
unsigned short int kpid;
int maxCPU, numLines;

char checkCount[MAX_BAD_PIDS];
char isAlive[MAX_BAD_PIDS];
unsigned short int badPids[MAX_BAD_PIDS];
char badUsers[MAX_BAD_PIDS][20];
char badProcs[MAX_BAD_PIDS][20];
int pp[2];

if(fork()) exit(0);


myPid=getpid();
myPpid=getppid();

for(i=0;i<MAX_BAD_PIDS;i++) badPids[i]=0;
openlog("monload2",0,LOG_DAEMON);

syslog(LOG_DAEMON|LOG_INFO,"User processes monitor started %i\n",myPid);
strcpy(line,"root ");
for(i=1;i<argc;i++) {strcat(line,argv[i]);strcat(line," ");}
syslog(LOG_DAEMON|LOG_INFO,"Ignoring processes owned by: %s\n",line);
syslog(LOG_DAEMON|LOG_INFO,"Check interval is %i seconds\n",CICLE_DELAY);
syslog(LOG_DAEMON|LOG_INFO,"Process CPU limit is %i%% at %i average system load (x100).\n",CPU_LEVEL1,SYS_LOAD_LEVEL1);
syslog(LOG_DAEMON|LOG_INFO,"Process CPU limit is %i%% at %i average system load (x100).\n",CPU_LEVEL2,SYS_LOAD_LEVEL2);
syslog(LOG_DAEMON|LOG_INFO,"Process MEM limit is %i%% at any average system load.\n",MEM_LIMIT);
syslog(LOG_DAEMON|LOG_INFO,"Process low priority timeout is %i seconds (%i checks)\n",COUNT_TO_NICE*CICLE_DELAY,COUNT_TO_NICE);
syslog(LOG_DAEMON|LOG_INFO,"Process send TERM timeout is %i seconds (%i checks)\n",COUNT_TO_TERM*CICLE_DELAY,COUNT_TO_TERM);
syslog(LOG_DAEMON|LOG_INFO,"Process send KILL timeout is %i seconds (%i checks)\n",COUNT_TO_KILL*CICLE_DELAY,COUNT_TO_KILL);
syslog(LOG_DAEMON|LOG_INFO,"MAJOR FAIL reboot timeout is %i seconds (%i checks)\n",MAIN_FAIL_REBOOT*CICLE_DELAY,MAIN_FAIL_REBOOT);
syslog(LOG_DAEMON|LOG_INFO,"MAJOR FAIL above %i average system load (x100), or on fatal error\n",MAIN_FAIL_LOAD_LIMIT);

mainFailCount=0;
signal(SIGALRM,stopPs);
signal(SIGCHLD,waitChild);


while(1)
   {
   psPid=0;
   reading=0;
   mainFail=0;
   maxCPU=100;

   alarm(CICLE_DELAY);
   /* get the load */
   if(pipe(pp)== -1) mainFail=1;
   if(!mainFail) psPid=fork();
   if(!mainFail)
	{
	if(psPid==-1) mainFail=1;
	else
		{
		if(psPid) close(pp[1]);
		else
			{
			close(pp[0]);dup2(pp[1],1);
			execl(CAT_EXEC,"cat","/proc/loadavg",0);
			exit(1);
			}
		}
	}
   if(!mainFail) 
	{
	reading=1;
	if(readLine(line,300,pp[0]))
		{
		aux=line;
		while(*aux!='.') aux++; aux[0]=aux[1];aux[1]=aux[2];aux[2]=0;
		sysLoad=atoi(line);
		if(sysLoad>MAIN_FAIL_LOAD_LIMIT) mainFail=1;
		if(sysLoad>=SYS_LOAD_LEVEL2) maxCPU=CPU_LEVEL2;
		else
		if(sysLoad<SYS_LOAD_LEVEL1) maxCPU=100;
		else
			{
			maxCPU=CPU_LEVEL2+(SYS_LOAD_LEVEL2-sysLoad)*kkkk/1000;
			}
		if(sysLoad>=SYS_LOAD_LEVEL1)
			syslog(LOG_DAEMON|LOG_INFO,"System load is %i (x100), CPU limit set to %i%% per process\n",sysLoad,maxCPU);
		}
	else mainFail=1;
	close(pp[0]);
	reading=0;
	}
   psPid=0;
	
   





   /* get the processes list */

   if(!mainFail)
	{
   	if(pipe(pp)== -1) mainFail=1;
	}

   if(!mainFail) psPid=fork();
   if(!mainFail)
	{
	if(psPid==-1) mainFail=1;
	else
		{
		if(psPid) close(pp[1]);
		else
			{
			close(pp[0]);dup2(pp[1],1);
			execl(PS_EXEC,"ps",PS_ARGS,0);
			exit(1);
			}
		}
	}
   

   if(!mainFail) 
	{
	reading=1;
        for(i=0;i<MAX_BAD_PIDS;i++) isAlive[i]=0;
	numLines=0;

	while(readLine(line,300,pp[0]))
		{
		aux=line;
		while(*aux > ' ') aux++;
		*aux=0;
		aux++;
		while(*aux <= ' ') aux++;
		aux1=aux+1;
		while(*aux1 > ' ') aux1++;
		*aux1=0;
		pid=atoi(aux);

		aux=aux1+1;
		while(*aux <= ' ') aux++;
		aux1=aux+1;
		while(*aux1>' ' && *aux1!='.') aux1++;
		*aux1=0;
		load=atoi(aux);

		aux=aux1+1;
		while(*aux <= ' ') aux++;
		aux1=aux+1;
		while(*aux1>' ' && *aux1!='.') aux1++;
		*aux1=0;
		mem=atoi(aux);


		if((load>=maxCPU) 
			|| mem>MEM_LIMIT) 
			{
			if(!CHECK_ROOT && !strcmp(line,"root")) issys=1; else issys=0;
			for(i=1;i<argc && !issys;i++) 
				if(!strcmp(argv[i],line)) {issys=1;break;}

			if(!issys) 
				{
				/*********** A BAD ONE **********************/	
				for(i=0;i<MAX_BAD_PIDS;i++) 
					if(pid==badPids[i]) break;

				if(i!=MAX_BAD_PIDS) /* you are still here */
					{
					isAlive[i]=1;
					checkCount[i]++;
					}
				else
					{ /* new one */
    					for(i=0;i<MAX_BAD_PIDS;i++) if(!badPids[i]) break;
					isAlive[i]=1;
					badPids[i]=pid;
					checkCount[i]=1;
					strcpy(badUsers[i],line);
					aux=aux1+1;
					while(*aux) aux++; aux-=3;
					while(*(aux-1)>32) aux--;
					aux1=badProcs[i];
					while(*aux>32) {*aux1=*aux;if((aux1-badProcs[i])<19) aux1++;aux++;} *aux1=0;
					}
		
				/*******************************************/
				}
			}
		numLines++;
		}
	close(pp[0]);
	reading=0;psPid=0;
	if(numLines<10) mainFail=1;


        for(i=0;i<MAX_BAD_PIDS;i++) if(!isAlive[i]) badPids[i]=0;
        for(i=0;i<MAX_BAD_PIDS;i++) 
	 if(badPids[i]) 
	  {
	  if(checkCount[i]>=COUNT_TO_KILL) 
		{syslog(LOG_DAEMON|LOG_WARNING,"Sending KILL to %s (%i) owned by %s - check n.%i\n",
			badProcs[i],badPids[i],badUsers[i],checkCount[i]);kill(badPids[i],SIGKILL);}
	  else
	  if(checkCount[i]>=COUNT_TO_TERM)
		{syslog(LOG_DAEMON|LOG_WARNING,"Sending TERM to %s (%i) owned by %s - check n.%i\n",
			badProcs[i],badPids[i],badUsers[i],checkCount[i]);kill(badPids[i],SIGTERM);}
	  else
	  if(checkCount[i]==COUNT_TO_NICE)
		{syslog(LOG_DAEMON|LOG_WARNING,"Setting %s (%i) to low priority, owned by %s - check n.%i\n",
			badProcs[i],badPids[i],badUsers[i],checkCount[i]);setpriority(PRIO_PROCESS,badPids[i],20);}
	  else
		{syslog(LOG_DAEMON|LOG_WARNING,"Bad process %s (%i) owned by %s - check n.%i\n",
			badProcs[i],badPids[i],badUsers[i],checkCount[i]);}
	  }


	}
  
    if(mainFail) {mainFailCount++;setpriority(PRIO_PROCESS,myPid,-20);}
    else {mainFailCount=0;setpriority(PRIO_PROCESS,myPid,0);}

    if(mainFailCount)
	{
	syslog(LOG_DAEMON|LOG_CRIT,"Major Fail detected, check %i/%i ", mainFailCount,MAIN_FAIL_REBOOT);
	if(mainFailCount>=MAIN_FAIL_REBOOT)
		{
		syslog(LOG_DAEMON|LOG_CRIT,"Forcing reboot");
		sync();sleep(1);
		syslog(LOG_DAEMON|LOG_CRIT,"Sending TERM to all above 100");
		for(kpid=100;kpid;kpid++) if(kpid!=myPid && kpid!=myPpid) kill(kpid,SIGTERM);
		syslog(LOG_DAEMON|LOG_CRIT,"Sending KILL to all above 100");
		sync();sleep(1);
		for(kpid=100;kpid;kpid++) if(kpid!=myPid && kpid!=myPpid) kill(kpid,SIGKILL);
		syslog(LOG_DAEMON|LOG_CRIT,"Executing reboot");
		sync();sleep(1);
		execl(REBOOT_EXEC,"reboot",0);
		syslog(LOG_DAEMON|LOG_CRIT,"Exec reboot failed, doing the hard way");
		fflush(stdout);
		reboot(RB_AUTOBOOT);
		sleep(1);
		exit(0);
		}
	else
		syslog(LOG_DAEMON|LOG_CRIT,"Hope this sorts out by itself");
	}
    fflush(stdout);
    for(i=0;i<=CICLE_DELAY;i++) sleep(1);
    }




}




